import emgdecompy.preprocessing
import emgdecompy.contrast
import emgdecompy.decomposition
import emgdecompy.viz
# read version from installed package
from importlib.metadata import version
__version__ = version("emgdecompy")